# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representation of python's std lib"""
import collections
from copy import deepcopy
from inspect import isclass
from typing import _GenericAlias
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.exception.type_check_error import (
    HashableError,
    KeyValueError,
    TupleArgLenError,
    TypeCheckError,
    UnRepresentedTypeError,
    WrongTypeError,
)
from adorn.unit.simple import Simple


class Python(Simple):
    """:class:`~adorn.unit.unit.Unit` that holds types in python's std lib."""

    _registry = collections.defaultdict(dict)

    def get(self, key: Any, orchestrator: "Orchestrator") -> Optional[type]:
        """Return the relevant python std lib type

        Args:
            key (Any): the type that may be part of python's
                std lib
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            Optional[type]: if ``key`` is part of python's std lib,
                then the type that is meant to represent the python
                std lib type will be returned, otherwise ``None``
        """
        if isclass(key) or isinstance(key, _GenericAlias):
            return self._get(key, orchestrator)
        return None

    def contains(self, obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the type is part of python's std lib.

        Args:
            obj (Any): the type that may be part of python's
                std lib
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            bool: if ``True``, the type is part of python's std lib
        """
        if not (isclass(obj) or isinstance(obj, _GenericAlias)):
            return False
        return any(self.relevant_types(obj, orchestrator))


class BuiltIn(Python):
    """Representation of python's built-in types"""

    base_type: type
    """the built-in python type the subclass is supposed to represent"""

    hashable: bool = True
    """python's built-in types are hashable"""

    unacceptable_subtype: Set[type] = {}
    """types that do not map to the
       given python built-in type.  This exists to prevent ``bool``
       from being confused as an ``int``."""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is part of python's built-in types.

        Args:
            obj (Type): the type that may be part of python's
                built-in types
            orchestrator (Orchestrator): container of all types, which
                is unused, since there are no nested types

        Returns:
            bool: if ``True``, the type is part of python's std lib
        """
        return (not isinstance(obj, _GenericAlias)) and (
            (issubclass(obj, cls.base_type) or (issubclass(obj, cls)))
            and obj not in cls.unacceptable_subtype
        )

    @classmethod
    def _from_obj(cls, target_cls: type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Instantiate an instance of the python built-in type

        Args:
            target_cls (type): The python built-in class, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is unused
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of the python built-in type, as specified
                by ``target_cls``
        """
        return cls.base_type(obj)

    @classmethod
    def additional_checks(
        cls, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Type checks specific to a given type

        Args:
            orchestrator (Orchestrator): container of all types,
                typically unused
            obj (Any): an instance that will have additional
                type checking performed against it

        Returns:
            Optional[TypeCheckError]: either a ``TypeCheckError`` containing
                the reason why the ``obj`` didn't pass the additional checks,
                otherwise ``None``
        """
        return None

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Check if an instance could be converted to the desired python built-in type

        Exceptions:
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` was not of the appropriate type

        Args:
            target_cls (Type): The python built-in class, ``obj``
                will be checked against
            orchestrator (Orchestrator): container of all types,
                which is unused
            obj (Any): an instance that will be checked against
                the type ``target_cls``

        Returns:
            Optional[TypeCheckError]: either a ``TypeCheckError`` containing
                the reason why the ``obj`` didn't type check, otherwise
                ``None``
        """
        if not isinstance(obj, cls.base_type):
            return WrongTypeError(target_cls=cls.base_type, obj=obj)
        return cls.additional_checks(orchestrator=orchestrator, obj=obj)


class Wrapper(Python):
    """Base class for python wrapper types."""

    @staticmethod
    def is_hashable(
        target_cls: Type,
        target_arg: Type,
        orchestrator: "Orchestrator",
        obj: Iterable[Any],
    ) -> Optional[TypeCheckError]:
        """Check if a given type may be hashed.

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                the nested argument, did not map to any of the types contained in
                the :class:`~adorn.orchestrator.orchestrator.Orchestrator`
            - :class:`~adorn.exception.type_check_error.HashableError`:
                the nested argument, could not be hashed

        Args:
            target_cls (Type): fully instantiated wrapper type
            target_arg (Type): nested type, that is required to be hashable
            orchestrator (Orchestrator): container of all types, used to recurse
            obj (Iterable[Any]): the instance, pre conversion, used to enrich
                error messages, if they occur

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue with the nested type, otherwise ``None``
        """
        if not orchestrator.contains(target_arg):
            return UnRepresentedTypeError(
                target_cls=target_cls, unit=orchestrator, obj=obj
            )
        unit = orchestrator.get(target_arg).get(target_arg, orchestrator)
        if not unit.hashable:
            return HashableError(target_cls=target_cls, unit=unit, obj=obj)


@Python.register("int")
class Int(BuiltIn):
    """Representation of python's ``int`` type"""

    base_type = int
    unacceptable_subtype: Set[type] = {bool}

    @classmethod
    def additional_checks(
        cls, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Type checks specific to ``int``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` was of type ``bool``, not ``int``

        Args:
            orchestrator (Orchestrator): container of all types,
                unused
            obj (Any): an instance that will have additional
                type checking performed against it

        Returns:
            Optional[TypeCheckError]: either a ``TypeCheckError`` containing
                the reason why the ``obj`` didn't pass the additional checks,
                otherwise ``None``
        """
        if type(obj) == bool:
            return WrongTypeError(int, obj)


@Python.register("str")
class Str(BuiltIn):
    """Representation of python's ``str`` type"""

    base_type = str


@Python.register("bool")
class Bool(BuiltIn):
    """Representation of python's ``bool`` type"""

    base_type = bool


@Python.register("float")
class Float(BuiltIn):
    """Representation of python's ``float`` type"""

    base_type = float


@Python.register("list")
class Rterable(Wrapper):
    """Representation of python's ``Iterable`` type"""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Iterable``

        Args:
            obj (Type): the type that may be ``Iterable``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested type

        Returns:
            bool: if ``True``, the type is ``Iterable``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return (
            origin in {collections.abc.Iterable, Iterable, List, list}
            and len(args) == 1
            and orchestrator.contains(args[0])
        )

    @classmethod
    def _from_obj(
        cls, target_cls: type, orchestrator: "Orchestrator", obj: Iterable[Any]
    ) -> Any:
        """Instantiate an instance of ``Iterable``

        Args:
            target_cls (type): The fully instantiated ``Iterable`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate the elements of the ``Iterable``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Iterable``, as specified
                by ``target_cls``
        """
        args = getattr(target_cls, "__args__", [])
        return [orchestrator.from_obj(args[0], i) for i in obj]

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Iterable[Any]
    ) -> Optional[TypeCheckError]:
        """Check if an instance could be converted to ``Iterable``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't be converted
                to the nested type the ``Iterable`` wraps
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` couldn't be converted to an ``Iterable``

        Args:
            target_cls (Type): The fully instantiated ``Iterable`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Iterable``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_arg = getattr(target_cls, "__args__", [])[0]
        if (not isinstance(obj, set)) and any(
            isinstance(obj, i) for i in [collections.abc.Iterable, Iterable, List, list]
        ):
            list_checks = (orchestrator.type_check(target_arg, i) for i in obj)
            relevant_checks = [
                (en, i) for en, i in enumerate(list_checks) if i is not None
            ]
            if relevant_checks:
                key_values = dict(relevant_checks)
                return KeyValueError(
                    target_cls=target_cls, key_values=key_values, obj=obj
                )
            return None
        return WrongTypeError(target_cls=target_cls, obj=obj)


@Python.register("set")
class Ret(Wrapper):
    """Represents Python's ``Set`` type"""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Set``

        Args:
            obj (Type): the type that may be ``Set``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested type

        Returns:
            bool: if ``True``, the type is ``Set``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return (
            origin in {Set, set} and len(args) == 1 and orchestrator.contains(args[0])
        )

    @classmethod
    def _from_obj(
        cls, target_cls: type, orchestrator: "Orchestrator", obj: Iterable[Any]
    ) -> Set[Any]:
        """Instantiate an instance of ``Set``

        Args:
            target_cls (type): The fully instantiated ``Set`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate the elements of the ``Set``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Set``, as specified
                by ``target_cls``
        """
        args = getattr(target_cls, "__args__", [])
        return {orchestrator.from_obj(args[0], i) for i in obj}

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Iterable[Any]
    ) -> Optional[str]:
        """Check if an instance could be converted to ``Set``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                the nested argument of ``Set``, did not map to any of the types
                contained in the :class:`~adorn.orchestrator.orchestrator.Orchestrator`
            - :class:`~adorn.exception.type_check_error.HashableError`:
                the nested argument of ``Set`` wasn't hashable
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't be converted
                to the nested type the ``Set`` wraps
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` couldn't be converted to a ``Set``

        Args:
            target_cls (Type): The fully instantiated ``Set`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Set``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_arg = getattr(target_cls, "__args__", [None])[0]
        if (target_arg is not None) and any(
            isinstance(obj, i) for i in [collections.abc.Iterable, Iterable, List, list]
        ):
            hashable_error = cls.is_hashable(
                target_cls=target_cls,
                target_arg=target_arg,
                orchestrator=orchestrator,
                obj=obj,
            )
            if hashable_error is not None:
                return hashable_error
            list_checks = (orchestrator.type_check(target_arg, i) for i in obj)
            relevant_checks = [i for i in list_checks if i is not None]
            if relevant_checks:
                key_values = dict(enumerate(relevant_checks))
                return KeyValueError(
                    target_cls=target_cls, key_values=key_values, obj=obj
                )
            return None
        return WrongTypeError(target_cls=target_cls, obj=obj)


@Python.register("tuple")
class Ruple(Wrapper):
    """Represents Python's ``Tuple`` type

    .. note::

        This representation of ``Tuple`` does not handle tuples of
        arbitrary length.  The full length of the ``Tuple`` must
        be specified in the type signature at instantiation time.
    """

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Tuple``

        Args:
            obj (Type): the type that may be ``Tuple``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested types

        Returns:
            bool: if ``True``, the type is ``Tuple``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return origin in {Tuple, tuple} and all(orchestrator.contains(i) for i in args)

    @classmethod
    def _from_obj(
        cls, target_cls: type, orchestrator: "Orchestrator", obj: Tuple[Any, ...]
    ) -> Any:
        """Instantiate an instance of ``Tuple``

        Args:
            target_cls (type): The fully instantiated ``Tuple`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate the elements of the ``Tuple``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Tuple``, as specified
                by ``target_cls``
        """
        args = getattr(target_cls, "__args__", [])
        return tuple(orchestrator.from_obj(i, j) for i, j in zip(args, obj))

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Tuple[Any, ...]
    ) -> Optional[str]:
        """Check if an instance could be converted to ``Tuple``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't be converted
                to the nested type the ``Tuple`` wraps
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` couldn't be converted to a ``Tuple``

        Args:
            target_cls (Type): The fully instantiated ``Tuple`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Tuple``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_args = getattr(target_cls, "__args__", [])
        if (
            not isinstance(obj, dict)
            and (target_args != [])
            and (not isinstance(obj, _GenericAlias))
            and isinstance(obj, Iterable)
        ):
            if len(target_args) != len(obj):
                return TupleArgLenError(target_cls, obj)
            list_checks = [
                orchestrator.type_check(i, j) for i, j in zip(target_args, obj)
            ]
            relevant_checks = [
                (en, i) for en, i in enumerate(list_checks) if i is not None
            ]
            if relevant_checks:
                key_values = dict(relevant_checks)
                return KeyValueError(
                    target_cls=target_cls, key_values=key_values, obj=obj
                )
            return None
        return WrongTypeError(target_cls=target_cls, obj=obj)


@Python.register("union")
class Rnion(Wrapper):
    """Represents Python's ``Union`` type"""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Union``

        Args:
            obj (Type): the type that may be ``Union``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested types

        Returns:
            bool: if ``True``, the type is ``Tuple``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return origin == Union and all(orchestrator.contains(i) for i in args)

    @classmethod
    def _from_obj(cls, target_cls: type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Instantiate an instance of ``Union``

        Args:
            target_cls (type): The fully instantiated ``Union`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate an element of the ``Union``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Union``, as specified
                by ``target_cls``

        Raises:
            KeyValueError: couldn't convert ``obj`` to any of the
                types specified in ``Union``
        """
        args = getattr(target_cls, "__args__", [])
        failed_args = []
        for arg in args:
            # do we really need to copy
            obj_prime = deepcopy(obj)
            output = orchestrator.type_check(arg, obj_prime)
            if output is None:
                return orchestrator.from_obj(arg, obj)
            else:
                failed_args.append((arg, output))

        key_values = {k: v for k, v in failed_args}
        raise KeyValueError(target_cls=target_cls, key_values=key_values, obj=obj)

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[str]:
        """Check if an instance could be converted to ``Union``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't be converted
                to a type specified in the ``Union``

        Args:
            target_cls (Type): The fully instantiated ``Union`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Union``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_args = getattr(target_cls, "__args__", [])
        list_checks = [(i, orchestrator.type_check(i, obj)) for i in target_args]
        if all(i is not None for _, i in list_checks):
            key_values = {k: v for k, v in list_checks}
            return KeyValueError(target_cls=target_cls, key_values=key_values, obj=obj)
        return None


@Python.register("dict")
class Rict(Wrapper):
    """Represents Python's ``Dict`` type"""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Dict``

        Args:
            obj (Type): the type that may be ``Dict``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested types

        Returns:
            bool: if ``True``, the type is ``Dict``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return (
            origin in {collections.abc.Mapping, Mapping, Dict, dict}
            and len(args) == 2
            and all(orchestrator.contains(i) for i in args)
        )

    @classmethod
    def _from_obj(
        cls, target_cls: type, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Dict[Any, Any]:
        """Instantiate an instance of ``Dict``

        Args:
            target_cls (type): The fully instantiated ``Dict`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate the keys and values
                of the ``Dict``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Dict``, as specified
                by ``target_cls``
        """
        args = getattr(target_cls, "__args__", [])
        return {
            orchestrator.from_obj(args[0], k): orchestrator.from_obj(args[1], v)
            for k, v in obj.items()
        }

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Optional[str]:
        """Check if an instance could be converted to ``Dict``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                the key type of ``Dict``, did not map to any of the types
                contained in the :class:`~adorn.orchestrator.orchestrator.Orchestrator`
            - :class:`~adorn.exception.type_check_error.HashableError`:
                the key type of ``Dict`` wasn't hashable
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't
                be converted to the key or value type the ``Dict`` wraps
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` couldn't be converted to a ``Dict``

        Args:
            target_cls (Type): The fully instantiated ``Dict`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Dict``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_args = getattr(target_cls, "__args__", [])
        if all(
            not isinstance(obj, i)
            for i in [collections.abc.Mapping, Mapping, Dict, dict]
        ):
            return WrongTypeError(target_cls=target_cls, obj=obj)
        hashable_error = cls.is_hashable(
            target_cls=target_cls,
            target_arg=target_args[0],
            orchestrator=orchestrator,
            obj=obj,
        )
        if hashable_error is not None:
            return hashable_error
        keys = []
        values = []
        for k, v in obj.items():
            # check key
            k_tc = orchestrator.type_check(target_args[0], k)
            if k_tc is not None:
                keys.append(k_tc)
            # check value
            v_tc = orchestrator.type_check(target_args[1], v)
            if v_tc:
                values.append(v_tc)

        if keys or values:
            # key(s) and/or value(s) didn't pass type_check
            key_values = dict()
            for en, (prefix, lst) in enumerate([("key", keys), ("value", values)]):
                for en_prime, i in enumerate(lst):
                    key_values[f"{prefix}_{target_args[en]}_{en_prime}"] = i
            return KeyValueError(target_cls=target_cls, key_values=key_values, obj=obj)
        return None



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Subclass of :class:`~adorn.unit.complex.Complex` to specify alternate construction of a class"""  # noqa: B950
from typing import Any
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.data.constructor import Constructor
from adorn.exception.type_check_error import TypeCheckError
from adorn.params import Params
from adorn.unit.complex import Complex


class Template(Complex):
    """Subclass of :class:`~adorn.unit.complex.Complex` to specify alternate construction of a class

    This construct allows for you to specify a different way of constructing
    an instance of :class:`~adorn.unit.complex.Complex`, without having
    to add an additional method to the given class of
    :class:`~adorn.unit.complex.Complex`.  This is especially useful
    for creating configurable constructs, that reduces parameters,
    while still covering most use cases.

    .. note

        The constructor for this class will be passed values from the configuration that
        haven't been instantiated.
    """  # noqa: B950

    @property
    def params(self) -> Params:
        """Configuration for the object being shadowed"""
        return self._params.duplicate()

    @property
    def _params(self) -> Params:  # pragma: no cover
        """Specifies a configuration for a class

        Raises:
            NotImplementedError: subclasses must override this method
        """
        raise NotImplementedError

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Check the arguments for the ``Template``

        Args:
            target_cls (Type): the target Type, the given ``Template`` will
                generate
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it
                can instantiate the ``Template``

        Returns:
            Optional[TypeCheckError]: Errors that would prevent you from instantiating
                the ``Template`` or the underlying ``target_cls`` that
                the ``Template`` generates an instance of, otherwise ``None``
        """  # noqa: B950
        return orchestrator.type_check(Constructor(cls), obj)

    @classmethod
    def _from_obj(cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Instantiate the ``Template``, then use :attr:`~adorn.unit.template.Template.params` to instantiate an instance of ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``Template`` will
                generate
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will instantiate an
                instance of ``Template``

        Returns:
            Any: an instance of ``target_cls``
        """  # noqa: B950, RST304
        return orchestrator.from_obj(
            target_cls, cls(**{k: v for k, v in obj.items() if k != "type"}).params
        )



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representation of ``dataclasses.dataclass``"""
from dataclasses import is_dataclass
from inspect import isclass
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Dict
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.data.constructor import Constructor
from adorn.exception.type_check_error import (
    ParamError,
    TypeCheckError,
    UnRepresentedTypeError,
)
from adorn.params import Params
from adorn.unit.unit import Unit


_D = TypeVar("_D")


class DataClass(Unit):
    """A wrapper around :meth:`~dataclasses.dataclass`

    .. code-block:: python

        from dataclasses import dataclass

        from adorn.orchestrator.base import Base
        from adorn.params import Params
        from adorn.unit.dataclass import DataClass

        class Example(DataClass):
            _registry = dict()

        @Example.register()
        @dataclass
        class A:
            a: int
            b: bool

        base = Base([Example()])
        assert base.from_obj(A, Params({"a": 1, "b": False})) == A(1, False)

    """  # noqa: RST304

    _registry: ClassVar[Dict[type, type]]

    @classmethod
    def register(cls) -> Callable[[Type[_D]], Type[_D]]:  # pragma: no cover
        """Add :meth:`~dataclasses.dataclass` to the :attr:`~adorn.unit.anum.Anum._registry`

        Returns:
            Callable[[Type[_D]], Type[_D]]: funcion that adds the decorated
                :meth:`~dataclasses.dataclass` to the ``_registry``
        """  # noqa: B950, D202, RST304

        def add_members(subclass: Type[_D]) -> Type[_D]:
            cls._registry[subclass] = subclass
            return subclass

        return add_members

    def get(
        self, key: Any, orchestrator: "Orchestrator"
    ) -> Optional[type]:  # pragma: no cover
        """Return the relevant ``Anum`` for the `key`

        Args:
            key (Any): the Type that may be a subclass of the ``Unit``
            orchestrator (Orchestrator): container of all types, not used

        Returns:
            Optional[type]: if ``key`` is an ``Anum`` then return it,
                otherwise ``None``
        """
        if self.contains(key, orchestrator):
            return key

    def contains(
        self, obj: Any, orchestrator: "Orchestrator"
    ) -> bool:  # pragma: no cover
        """Check if `obj` is registered to ``DataClass``

        Args:
            obj (Any): the Type that may be a subclass of ``Anum``
            orchestrator (Orchestrator): container of all types, not used

        Returns:
            bool: if ``True`` then ``obj`` is represented by ``Anum``,
                otherwise ``False``
        """
        return isclass(obj) and is_dataclass(obj) and obj in self._registry

    def type_check(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:  # pragma: no cover
        """Ensure `obj` can be converted to the type ``target_cls``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``
            - :class:`~adorn.exception.type_check_error.ParamError`:
                ``obj`` was not of type :class:`~adorn.params.Params`

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, used to check the
                arguments of the constructor for the given ``dataclass``
            obj (Any): a :class:`~adorn.params.Params`, that will be checked to
                see if it can be converted to an instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        if not self.contains(target_cls, orchestrator):
            return UnRepresentedTypeError(target_cls=target_cls, unit=self, obj=obj)

        if not isinstance(obj, Params):
            return ParamError(target_cls=target_cls, obj=obj)

        return orchestrator.type_check(Constructor(target_cls), obj)

    def from_obj(self, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> _D:
        """Convert `obj` to an instance of the type ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into
            orchestrator (Orchestrator): container of all types, used to instantiate
                the arguments for the constructor of the given ``dataclass``
            obj (Any): an instance, that will be converted to an
                instance of ``target_cls``

        Returns:
            _D: an instantiated :meth:`~dataclasses.dataclass`
        """
        return orchestrator.from_obj(Constructor(target_cls), obj)



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Container of types with a class hierarchy."""
import inspect
import logging
from collections import defaultdict
from typing import _GenericAlias
from typing import Any
from typing import Callable
from typing import cast
from typing import ClassVar
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.data.constructor import Constructor
from adorn.exception.configuration_error import ConfigurationError
from adorn.exception.type_check_error import (
    ComplexTypeMismatchError,
    ParamError,
    TypeCheckError,
    UnRepresentedTypeError,
)
from adorn.params import Params
from adorn.unit.unit import Unit

logger = logging.getLogger(__name__)

_T = TypeVar("_T")
_UnitT = TypeVar("_UnitT", bound="Unit")
_ComplexT = TypeVar("_ComplexT", bound="Complex")
_SubclassRegistry = Dict[str, Tuple[type, Optional[str]]]


class Complex(Unit):
    """:class:`~adorn.unit.unit.Unit` that preserves class hierarchy

    It is recommended you use this construct for your custom package.
    """

    _registry: ClassVar[DefaultDict[type, _SubclassRegistry]]
    """maps a parent class to all of its subclasses and their associated names"""
    _subclass_registry: Dict[type, Tuple[type, str]]
    """classes that can be instantiated, mapped to their parent and name"""
    _intermediate_registry: DefaultDict[type, List[type]]
    """mapping of classes, who are parents, to their parents"""
    _parent_registry: DefaultDict[_ComplexT, List[_ComplexT]]
    """mapping of parents to their children, who are also parents"""

    constructor_type: Optional[str] = None
    """The name of the :class:`~adorn.unit.constructor_value.ConstructorValue`
       you want to instantiate the arguments for your class"""
    parameter_order: Optional[List[str]] = None
    """The order in which the
       parameters for a class should be instantiated"""
    constructor_name: Optional[str] = None
    """Name of the constructor to use to instantiate the class"""

    @classmethod
    def root(cls: Type[_ComplexT]) -> Callable[[Type[_T]], Type[_T]]:
        """Initialize all the class attributes needed to track the class hieracrchy

        This decorator is typically called on a direct descendant of
        :class:`~adorn.unit.complex.Complex` enabling all of the direct
        descendant's children to be tracked.

        Returns:
            Callable[[Type[_T]], Type[_T]]: decorator that initializes the
                class attributes for tracking
        """

        def _root(subclass: Type[_T]) -> Type[_T]:
            subclass._registry = defaultdict(dict)
            subclass._subclass_registry = dict()
            subclass._intermediate_registry = defaultdict(list)
            subclass._parent_registry = defaultdict(list)
            return subclass

        return _root

    @classmethod
    def register(
        cls: Type[_ComplexT], name: Optional[str]
    ) -> Callable[[Type[_T]], Type[_T]]:
        """Decorator that adds a class to the class hierarchy

        Args:
            name (Optional[str]): The ``str`` that will map to the given type, if
                ``None``, then the class is both a child of another ``Complex`` type
                and a parent of other ``Complex`` type(s), but cannot be
                instantiated

        Returns:
            Callable[[Type[_T]], Type[_T]]: funcion that adds the decorated type
                to the ``_registry`` with the ``name`` being the key and the decorated
                class being the value
        """
        registry = cls._registry[cls]

        def add_subclass_to_registry(subclass: Type[_T]) -> Type[_T]:
            # Add to registry, raise an error if key has already been used.
            if name is not None:
                if name in registry:
                    message = (
                        f"Cannot register {name} as {cls.__name__}; "
                        f"name already in use for {registry[name].__name__}"
                    )
                    raise ConfigurationError(message)
                registry[name] = subclass
                cls._subclass_registry[subclass] = (cls, name)
                if cls in cls._subclass_registry:
                    parent, _ = cls._subclass_registry[cls]
                    cls._parent_registry[parent].append(cls)
            else:
                cls._intermediate_registry[subclass].append(cls)
                cls._parent_registry[cls].append(subclass)
            return subclass

        return add_subclass_to_registry

    @classmethod
    def get_instantiate_children(cls: Type[_ComplexT]) -> Dict[str, _ComplexT]:
        """Get the subclasses for a class that can be instantiated

        Returns:
            Dict[str, _ComplexT]: All the subclasses of a class that can be instantiated
        """
        instantiate_children = cls._explode_registry()
        for i in cls.get_intermediate_children():
            instantiate_children = {
                **i.get_instantiate_children(),
                **instantiate_children,
                **i._explode_registry(),
            }
        return instantiate_children

    @classmethod
    def _explode_registry(cls: Type[_ComplexT]) -> Dict[str, _ComplexT]:
        """Helper method, to get all the descendants of a class

        Returns:
            Dict[str, _ComplexT]: Dictionary containing all the descendants of a class
        """
        instantiate_children = (
            dict() if cls not in cls._subclass_registry else {cls.name(): cls}
        )
        for v in cls._registry[cls].values():
            instantiate_children = {
                **instantiate_children,
                **v.get_instantiate_children(),
            }
        return instantiate_children

    @classmethod
    def get_direct_parent(
        cls: Type[_ComplexT],
    ) -> Optional[Union[_ComplexT, List[_ComplexT]]]:
        """Get the parent(s) of a class

        Returns:
            Optional[Union[_ComplexT, List[_ComplexT]]]: if ``None``, then the class
                has no parents, if ``_ComplexT`` then the class has a parent,
                if ``List[_ComplexT]`` then the class has multiple parents
        """
        if cls in cls._parent_registry:
            return None

        if cls in cls._intermediate_registry:
            return cls._intermediate_registry[cls]

        return cls._subclass_registry[cls][0]

    @classmethod
    def get_intermediate_children(cls: Type[_ComplexT]) -> List[_ComplexT]:
        """Get the children of a class, that are also parents

        .. note::
            Some of the classes in the list may not be able to be instantiated

        Returns:
            List[_ComplexT]: list of subclasses of that are also parents
        """
        return cls._parent_registry.get(cls, [])

    @classmethod
    def name(cls: Type[_ComplexT]) -> str:
        """Get the string associated with a given class

        Returns:
            str: the string associated with a class

        Raises:
            Exception: the given class, cannot be instantiated and therefore
                does not have a name
        """
        if cls not in cls._subclass_registry:
            raise Exception(f"{cls} was not registered as a constructable subclass")
        return cls._subclass_registry[cls][1]

    @classmethod
    def by_name(
        cls: Type[_ComplexT], name: Optional[str] = None
    ) -> Callable[..., _ComplexT]:
        """Returns a callable function that constructs an argument of the class.

        Args:
            name (Optional[str]): name of the class you want the constructor for,
                if ``None``, :meth:`~adorn.unit.complex.Complex.name` will be used

        Returns:
            Callable[..., _ComplexT]: constructor for the requested class
        """
        if cls not in cls._subclass_registry:
            return cls
        name = name or cls.name()
        logger.debug(f"instantiating registered subclass {name} of {cls}")
        subclass = cls.resolve_class_name(name)
        return cast(Type[_ComplexT], subclass)

    @classmethod
    def resolve_class_name(
        cls: Type[_ComplexT], name: Optional[str] = None
    ) -> Type[_ComplexT]:
        """Returns the class that corresponds to the given `name`

        Args:
            name (str): name of the class the caller wants

        Returns:
            Type[_ComplexT]: the type associated with ``name`` specified
        """
        name = name or cls.name()
        children = cls.get_instantiate_children()
        return children[name]

    @classmethod
    def list_available(cls: Type[_ComplexT]) -> List[str]:
        """List all the names of subclasses that can be instantiated

        Returns:
            List[str]: names of subclasses that can be instantiated
        """
        return list(cls.get_instantiate_children().values())

    def match(self, key: Any) -> bool:
        """Check if the key is a subclass of ``Complex``

        Args:
            key (Any): potential type to be checked

        Returns:
            bool: if ``True`` then key is a Type of the appropriate type
        """
        return (
            (not isinstance(key, _GenericAlias))
            and inspect.isclass(key)
            and issubclass(key, type(self))
        )

    def get(self, key: Any, orchestrator: "Orchestrator") -> Optional[type]:
        """Return the relevant registered type

        Args:
            key (Any): the type that may be a registered value
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            Optional[type]: if ``key`` is a registered value,
                then the registered value is returned, otherwise ``None``
        """
        if self.match(key):
            return key._get()
        return None

    @classmethod
    def _get(cls: Type[_ComplexT]) -> Optional[type]:
        """Get the relevant class

        This method exists to faciliate potential future use cases
        that may require additional logic for getting the correct class

        Returns:
            Optional[type]: return the class associated with the classmethod
        """
        return cls

    def contains(self, obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` is registered

        Args:
            obj (Any): potentially a type that may be registered
            orchestrator (Orchestrator): container of all types, which
                is used for nested types

        Returns:
            bool: if ``True``, the type is registered
        """
        if self.match(obj):
            return obj._contains(obj, orchestrator)
        return False

    @classmethod
    def _contains(cls: Type[_ComplexT], obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the ``cls`` has been registered

        Args:
            obj (Any): this is unused
            orchestrator (Orchestrator): this is unused

        Returns:
            bool: if ``True``, the type can be represented by ``cls``
        """
        return cls in cls._subclass_registry or cls in cls._registry

    def general_check(
        self,
        cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
        is_from_obj: bool = False,
    ) -> Optional[TypeCheckError]:
        """Checks that need to be performed for both ``type_check`` and ``from_obj``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``
            - :class:`~adorn.exception.type_check_error.ParamError`:
                ``obj`` was not of type :class:`~adorn.params.Params`
            - :class:`~adorn.exception.type_check_error.ComplexTypeMismatchError`:
                The ``type`` key contained in ``obj`` had a value that did not
                map to any of the valid subclasses of ``target_cls``

        Args:
            cls (Type): the target class that ``obj`` is trying to be converted
                into
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): the instance that will have the sanity checks applied
                to it
            is_from_obj (bool): if ``True`` this check is being done for
                :meth:`~adorn.unit.complex.Complex.from_obj`, which allows
                instances of ``cls`` to be considered valid

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        if not self.contains(cls, orchestrator):
            return UnRepresentedTypeError(cls, self, obj)
        if is_from_obj and isinstance(obj, cls):
            return obj
        if not isinstance(obj, Params):
            return ParamError(target_cls=cls, obj=obj)

        children = cls.get_instantiate_children()
        if obj.get("type") not in children:
            return ComplexTypeMismatchError(cls, children, obj)

    def type_check(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Ensure `obj` can be converted to the type ``target_cls``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``
            - :class:`~adorn.exception.type_check_error.ParamError`:
                ``obj`` was not of type :class:`~adorn.params.Params`
            - :class:`~adorn.exception.type_check_error.ComplexTypeMismatchError`:
                The ``type`` key contained in ``obj`` had a value that did not
                map to any of the valid subclasses of ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, usually of type :class:`~adorn.params.Params`,
                that will be checked to see if it can be converted to an
                instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        general_check_output = self.general_check(target_cls, orchestrator, obj)
        if general_check_output is not None:
            return general_check_output
        ru = target_cls.resolve_class_name(obj.get("type"))
        return ru._type_check(target_cls, orchestrator, obj)

    @classmethod
    def _type_check(
        cls: Type[_ComplexT], target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Check that the args, specified in ``obj``, for ``target_cls`` type check

        The checking of arguments for a ``target_cls`` is delegated to
        :class:`~adorn.unit.constructor_value.ConstructorValue`

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, usually of type :class:`~adorn.params.Params`,
                that will be checked to see if it can be converted to an
                instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        return orchestrator.type_check(Constructor(cls), obj)

    def from_obj(self, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Convert `obj` to an instance of ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, usually of type :class:`~adorn.params.Params`,
                that will be converted into an instance of ``target_cls``

        Returns:
            Any: an instance of ``target_cls`` derived from ``obj``

        Raises:
            UnRepresentedTypeError: ``obj`` did not map to any of the types contained in
                the ``_registry``
            ParamError: ``obj`` was not of type :class:`~adorn.params.Params`
            ComplexTypeMismatchError: The ``type`` key contained in ``obj`` had a value
                that did not map to any of the valid subclasses of ``target_cls``
        """  # noqa: DAR401, DAR402
        general_check_output = self.general_check(
            target_cls, orchestrator, obj, is_from_obj=True
        )
        if isinstance(general_check_output, TypeCheckError):
            raise general_check_output
        if general_check_output is not None:
            return general_check_output

        ru = target_cls.resolve_class_name(obj.get("type"))
        return ru._from_obj(target_cls, orchestrator, obj)

    @classmethod
    def _from_obj(
        cls: Type[_ComplexT], target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Any:
        """Convert the args, specified in ``obj``, to instances to be passed ``target_cls`` constructor

        The instantiation of arguments for the``target_cls`` and instantiation of the ``target_cls``
        are delegated to :class:`~adorn.unit.constructor_value.ConstructorValue`

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, usually of type :class:`~adorn.params.Params`,
                that will be converted into an instance of ``target_cls``

        Returns:
            Any: an instance of ``target_cls`` derived from ``obj``
        """  # noqa: B950
        if isinstance(obj, cls):
            return obj
        return orchestrator.from_obj(Constructor(cls), obj)



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representation of instances, specifically ``None`` and ``typing.Any``"""
from collections import defaultdict
from typing import Any
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.exception.type_check_error import TypeCheckError, WrongTypeError
from adorn.unit.simple import Simple


class Value(Simple):
    """:class:`~adorn.unit.unit.Unit` that holds instances

    This class exists to handle ``None``, which is allowed to
    be used as a type and a value, as well as ``typing.Any``.
    """

    _registry = defaultdict(dict)

    def get(self, key: Any, orchestrator: "Orchestrator") -> Optional[Type]:
        """Return the relevant type that represents a value

        Args:
            key (Any): the type or instance that may be part of ``Value``
            orchestrator (Orchestrator): container of all types, typically
                unused

        Returns:
            Optional[Type]: if ``key`` is represented by ``Value``,
                then the type that is meant to represent an instance
                will be returned, otherwise ``None``
        """
        return self._get(key, orchestrator)

    def contains(self, obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the type or instance is part of ``Value``

        Args:
            obj (Any): the type or instance that may be part of ``Value``
            orchestrator (Orchestrator): container of all types, typically
                unused

        Returns:
            bool: if ``True``, the type or instance is contained in ``Value``
        """
        return any(self.relevant_types(obj, orchestrator))


@Value.register("none")
class Rone(Value):
    """Representation of python's `NoneType`, aka ``None``"""

    NoneType = type(None)

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type or instance is ``None``

        Args:
            obj (Type): the type or instance that may be ``None``
            orchestrator (Orchestrator): container of all types, which
                is unused, since there are no nested types

        Returns:
            bool: if ``True``, ``obj`` was ``None`` or ``NoneType``
        """
        return obj is None or obj == cls.NoneType

    @classmethod
    def _from_obj(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> None:
        """Generate an instance of ``None``

        Args:
            target_cls (Type): This is ignored
            orchestrator (Orchestrator): This is ignored
            obj (Any): This is ignored

        Returns:
            None: an instance of ``None`` is returned
        """
        return None

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Check if ``obj`` is ``None``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` was not ``None``

        Args:
            target_cls (Type): This is ignored
            orchestrator (Orchestrator): This is ignored
            obj (Any): instance that may be ``None``

        Returns:
            Optional[TypeCheckError]: if ``obj`` wasn't ``None`` an
                exception will be returned, otherwise ``None``
        """
        if not cls._contains(obj, orchestrator):
            return WrongTypeError(cls.NoneType, obj)


@Value.register("any")
class Nny(Value):
    """Representation of python's ``typing.Any``"""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type or instance is ``typing.Any``

        Args:
            obj (Type): the type or instance that may be ``typing.Any``
            orchestrator (Orchestrator): container of all types, which
                is unused, since there are no nested types

        Returns:
            bool: if ``True``, ``obj`` was ``typing.Any``
        """
        return obj == Any

    @classmethod
    def _from_obj(cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Generate an instance of ``typing.Any``

        Args:
            target_cls (Type): This is ignored
            orchestrator (Orchestrator): This is ignored
            obj (Any): This is returned

        Returns:
            Any: return the ``obj`` that was passed
        """
        return obj

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Check if ``obj`` is ``typing.Any``.

        This will never produce an exception, since any object in python
        is always a ``typing.Any``

        Args:
            target_cls (Type): This is ignored
            orchestrator (Orchestrator): This is ignored
            obj (Any): This is ignored

        Returns:
            Optional[TypeCheckError]: always ``None``
        """
        return None



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Collection of similar types or values."""
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.exception.type_check_error import TypeCheckError

_T = TypeVar("_T")
A = TypeVar("A")


class Unit:
    """A collection of types or values, that usually have some shared characteristic(s)."""  # noqa: B950

    hashable: bool = False
    """If ``True`` then instances of the given type are hashable"""

    @classmethod
    def register(cls, name: str) -> Callable[[Type[_T]], Type[_T]]:  # pragma: no cover
        """Decorator that maps a string to the type

        Args:
            name (str): a string that will be the key for a ``Type``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def get(
        self, key: Any, orchestrator: "Orchestrator"
    ) -> Optional[type]:  # pragma: no cover
        """Return the relevant ``Type`` for the `key`

        Args:
            key (Any): the Type that may be a subclass of the ``Unit``
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def contains(
        self, obj: Any, orchestrator: "Orchestrator"
    ) -> bool:  # pragma: no cover
        """Check if `obj` has a corresponding type in `Unit`

        Args:
            obj (Any): the Type that may be a subclass of the given ``Unit``
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def type_check(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:  # pragma: no cover
        """Ensure `obj` can be converted to the type ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it
                can be converted to an instance of ``target_cls``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def from_obj(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Any:  # pragma: no cover
        """Convert `obj` to an instance of the type ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be converted to an
                instance of ``target_cls``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Objects that enable grid searching adorn configs"""
from functools import reduce
from itertools import chain
from itertools import groupby
from itertools import product
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

from adorn.params import Params
from adorn.unit.complex import Complex
from adorn.unit.search.search import SearchElement
from adorn.unit.search.search import SearchSpace
from adorn.unit.template import Template


GRID_OUTPUT = TypeVar("GRID_OUTPUT", List[Params], List[List[Params]])


@SearchElement.root()
class GridElement(SearchElement):
    """Individual component that make up a space to search over"""

    def __init__(self) -> None:
        super().__init__()

    def __len__(self) -> int:  # pragma: no cover
        """Number of elements the component is adding to the search space"""
        raise NotImplementedError

    def __call__(self) -> List[Dict[str, Any]]:  # pragma: no cover
        """Generate the collection of values to add to the search space

        Raises:
            NotImplementedError: needs to be implemented by child class
        """
        raise NotImplementedError

    @staticmethod
    def str_to_list(str_or_list: Union[str, List[str]]) -> List[str]:
        """Convert a string to a list

        Args:
            str_or_list (Union[str, List[str]]): a string to be converted
                to a list or a list that will have nothing done to it

        Returns:
            List[str]: a string that has been promoted to a list or just
                the original list that was passed
        """
        if isinstance(str_or_list, str):
            return [str_or_list]
        return str_or_list


class GridElementTemplate(GridElement, Template):
    """A ``Template`` for components of a grid search"""


@GridElement.register("base_grid")
class BaseGrid(GridElement):
    """Generate a dict to be added to the search space for each value

    Args:
        keys (Union[str, List[str]]): name(s) to assign to each search value
        values (List[Any]): collection of values to search over
    """

    def __init__(self, keys: Union[str, List[str]], values: List[Any]) -> None:
        super().__init__()
        self.keys = self.str_to_list(keys)
        self.values = values

    def __len__(self) -> int:
        """Number of elements added to the search space"""
        return len(self.values)

    def __call__(self) -> List[Dict[str, Any]]:
        """Create a ``Dict`` for each search value to be added to the search space

        Returns:
            List[Dict[str, Any]]: a collection of search values to
                be added to the search space
        """
        return [{k: v for k in self.keys} for v in self.values]


@BaseGrid.register("id_grid")
class IdGrid(GridElementTemplate):
    """Add a single value to a grid search

    This is useful when you want to override a value for a
    given search space
    """

    def __init__(self, keys: Union[str, List[str]], value: Any) -> None:
        super().__init__()
        self.keys = keys
        self.value = value

    @property
    def _params(self) -> Params:
        """Config to create the search component"""
        return Params({"type": "base_grid", "keys": self.keys, "values": [self.value]})


@SearchSpace.root()
class GridSearch(SearchSpace):
    """Generate a grid search space by composing multiple grid components"""

    def __init__(self) -> None:
        super().__init__()

    def __call__(self) -> List[Dict[str, Any]]:  # pragma: no cover
        """Generate all elements of the search space

        Raises:
            NotImplementedError: needs to be implemented by child class
        """
        raise NotImplementedError

    @staticmethod
    def tuple_to_dict(tpl: Tuple[Dict[str, Any], ...]) -> Dict[str, Any]:
        """Convert a tuple of dictionaries to a single dictionary

        Args:
            tpl (Tuple[Dict[str, Any], ...]): collection of dictionaries

        Returns:
            Dict[str, Any]: a dictionary made of all the dictionaries
                passed to this method
        """
        return dict(chain(*map(lambda i: i.items(), tpl)))


@GridSearch.register("base_grid_search")
class BaseGridSearch(GridSearch):
    """Generate a search space by taking the ``itertools.product`` of other search spaces and components

    Args:
        search_space (List[Union[GridElement, GridSearch]]): collection of grid
            objects to be composed
    """  # noqa: B950

    def __init__(self, search_space: List[Union[GridElement, GridSearch]]) -> None:
        super().__init__()
        self.search_space = search_space

    def __len__(self) -> int:
        """Number of elements to be searched over"""
        lengths = map(lambda i: len(i), self.search_space)
        return reduce(lambda i, j: i * j, lengths, 1)

    def __call__(self) -> List[Dict[str, Any]]:
        """Take the ``itertools.product`` of the provided :attr:`~adorn.unit.search.grid.BaseGridSearch.search_space` to generate a new grid search space

        Returns:
            List[Dict[str, Any]]: grid search space made of the ``itertools.product`` of
                grid elements and search spaces
        """  # noqa: B950, RST304
        iter_of_tpl = product(*[i() for i in self.search_space])
        return [self.tuple_to_dict(tpl) for tpl in iter_of_tpl]


@GridSearch.register("list_grid_search")
class ListGridSearch(GridSearch):
    """Concatenate grid search spaces to create a single search space

    Args:
        search_space (List[GridSearch]): multiple search spaces to
            be concatenated
    """

    def __init__(self, search_space: List[GridSearch]) -> None:
        super().__init__()
        self.search_space = search_space

    def __len__(self) -> int:
        """Number of elements to be searched over"""
        lengths = map(lambda i: len(i), self.search_space)
        return reduce(lambda i, j: i + j, lengths, 0)

    def __call__(self) -> List[Dict[str, Any]]:
        """Grid search space made up of a concatenation of search spaces

        Returns:
            List[Dict[str, Any]]: a collection of search space elements
        """
        return [j for i in self.search_space for j in i()]


@GridSearch.register("file_grid_search")
class FileGridSearch(GridSearch):
    """A grid search space made of a collection of adorn configs

    Args:
        filenames (List[str]): location of the adorn configs on disk
    """

    def __init__(self, filenames: List[str]) -> None:
        super().__init__()
        self.filenames = filenames

    def __len__(self) -> int:
        """Number of elements to be searched over"""
        return len(self.filenames)

    def __call__(self) -> List[Dict[str, Any]]:
        """Grid search space made up of serialized adorn configs

        Returns:
            List[Dict[str, Any]]: a collection of search space elements
                where each element was an adorn config read from disk
        """
        return [Params.from_file(i).as_dict() for i in self.filenames]


@Complex.root()
class Organize(Complex):
    """Alter the order of a list of grid points"""

    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self, search_list: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:  # pragma: no cover
        """Rearrange a list of grid points

        Args:
            search_list (List[Dict[str, Any]]): list of grid points

        Raises:
            NotImplementedError: needs to be implemented by child class
        """
        raise NotImplementedError


@Organize.register("id_organize")
class IDOrganize(Organize):
    """Perform no alterations to the list of grid points"""

    def __call__(self, search_list: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Return the list of grid points without alterations

        Args:
            search_list (List[Dict[str, Any]]): list of grid points

        Returns:
            List[Dict[str, Any]]: list of grid points passed, without
                alterations
        """
        return search_list


@Organize.register("sort_organize")
class SortOrganize(Organize):
    """Sort a list of grid points based on the values contained in each element

    Args:
        keys (Union[str, List[str]]): names of the values to sort the values
            of a list
    """

    def __init__(self, keys: Union[str, List[str]]) -> None:
        super().__init__()
        self.keys = keys

    def generate_key(self, dct: Dict[str, Any]) -> Union[Any, Tuple[Any, ...]]:
        """Get the values associated with a grid point for sorting

        Args:
            dct (Dict[str, Any]): the grid point, which will have values parsed from it

        Returns:
            Union[Any, Tuple[Any, ...]]: values from a grid point to be used for sorting
        """
        return tuple(
            dct[key]
            for key in (self.keys if isinstance(self.keys, list) else [self.keys])
        )

    def __call__(self, search_list: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Sort a list of grid points based on a subset of values

        Args:
            search_list (List[Dict[str, Any]]): collection of grid points to be sorted

        Returns:
            List[Dict[str, Any]]: list of grid points passed, sorted
                based on a subset of their values
        """
        return sorted(search_list, key=self.generate_key)


@Complex.root()
class Group(Complex):
    """Partition a list of grid points based on shared characteristics"""

    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self, search_list: List[Dict[str, Any]]
    ) -> List[List[Dict[str, Any]]]:  # pragma: no cover
        """Partition a list of grid points

        Args:
            search_list (List[Dict[str, Any]]): collection of grid
                points to be partitioned

        Raises:
            NotImplementedError: needs to be implemented by child class
        """
        raise NotImplementedError


@Group.register("id_group")
class IDGroup(Group):
    """Add each element of a list of grid points to its own partition"""

    def __init__(self) -> None:
        super().__init__()

    def __call__(self, search_list: List[Dict[str, Any]]) -> List[List[Dict[str, Any]]]:
        """Place each element of the grid list into its own group

        Args:
            search_list (List[Dict[str, Any]]): collection of grid points

        Returns:
            List[List[Dict[str, Any]]]: each element is in its own partition
        """
        return [[i] for i in search_list]


@Group.register("group_by")
class GroupBy(Group):
    """Partition a list of grid points based on common values

    Args:
        keys (Union[str, List[str]]): names of the values to group
            the grid elements by
    """

    def __init__(self, keys: Union[str, List[str]]) -> None:
        super().__init__()
        self.sort_organize = SortOrganize(keys=keys)

    def __call__(self, search_list: List[Dict[str, Any]]) -> List[List[Dict[str, Any]]]:
        """Group grid points based on common values

        Args:
            search_list (List[Dict[str, Any]]): collection of grid points

        Returns:
            List[List[Dict[str, Any]]]: partitions of elements with common values
        """
        return [
            list(i)
            for _, i in groupby(
                self.sort_organize(search_list=search_list),
                key=self.sort_organize.generate_key,
            )
        ]


@Complex.root()
class GridOrchestrator(Complex, Generic[GRID_OUTPUT]):
    """Generate a grid search space and optionally alter the space"""

    constructor_name = "__init__"

    def __init__(self) -> None:
        super().__init__()

    def __call__(self) -> GRID_OUTPUT:  # pragma: no cover
        """Produce a grid search space of collection of grid search spaces

        Raises:
            NotImplementedError: needs to be implemented by child class
        """
        raise NotImplementedError


@GridOrchestrator.register("grid_orch_list")
class GridList(GridOrchestrator[List[Params]]):
    """Generate a grid search space and optionally alter the order of the space

    Args:
        grid_search (GridSearch): specification of a grid search space
        organize (Optional[Organize]): optionally alter the order of a search
            space, if ``None`` the order is not altered
    """

    def __init__(
        self, grid_search: GridSearch, organize: Optional[Organize] = None
    ) -> None:
        super().__init__()
        self.grid_search = grid_search
        self.organize = organize or IDOrganize()

    def __call__(self) -> List[Params]:
        """Generate a search space and order it

        Returns:
            List[Params]: a grid search space that has potentially had
                its order altered
        """
        return self.organize(self.grid_search())


@GridOrchestrator.register("grid_orch_nested_list")
class GridNestedList(GridOrchestrator[List[List[Params]]]):
    """Generate a collection of grid search spaces

    Args:
        grid_search (GridSearch): specification of a grid search space
        group (Optional[Group]): optional specification of how to create
            grid search spaces, if ``None`` each element in the grid search
            becomes its own grid search space
    """

    def __init__(self, grid_search: GridSearch, group: Optional[Group] = None) -> None:
        super().__init__()
        self.grid_search = grid_search
        self.group = group or IDGroup()

    def __call__(self) -> List[List[Params]]:
        """Create a collection of grid search spaces

        Returns:
            List[List[Params]]: a collection of grid search spaces
        """
        return self.group(self.grid_search())



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Container of types with a flat class hierarchy."""
import logging
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import DefaultDict
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.exception.configuration_error import ConfigurationError
from adorn.exception.type_check_error import TypeCheckError, UnRepresentedTypeError
from adorn.unit.unit import Unit

logger = logging.getLogger(__name__)

_T = TypeVar("_T")
A = TypeVar("A")
_SubclassRegistry = Dict[str, type]


class Simple(Unit):
    """:class:`~adorn.unit.unit.Unit` for types with a flat class hierarchy

    Attributes:
        _registry (ClassVar[DefaultDict[type, _SubclassRegistry]]): contains mapping
            from ``str`` to ``Type``
    """

    _registry: ClassVar[DefaultDict[type, _SubclassRegistry]]

    @classmethod
    def register(cls, name: str) -> Callable[[Type[_T]], Type[_T]]:
        """Decorator that adds a type to the ``_registry``

        Args:
            name (str): The ``str`` that will map to the given type

        Returns:
            Callable[[Type[_T]], Type[_T]]: funcion that adds the decorated type
                to the ``_registry`` with the ``name`` being the key and the decorated
                class being the value
        """
        registry = cls._registry[cls]

        def add_subclass_to_registry(subclass: Type[_T]) -> Type[_T]:
            # Add to registry, raise an error if key has already been used.
            if name in registry:
                message = (
                    f"Cannot register {name} as {cls.__name__}; "
                    f"name already in use for {registry[name].__name__}"
                )
                raise ConfigurationError(message)
            registry[name] = subclass
            return subclass

        return add_subclass_to_registry

    def _get(self, key: type, orchestrator: "Orchestrator") -> Optional[type]:
        """Return the relevant type from the ``_registry``

        Args:
            key (type): the type that may be a value in the
                ``_registry``
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            Optional[type]: if ``key`` is a value in the ``_registry``,
                then the value is returned, otherwise ``None``
        """
        try:
            return next(self.relevant_types(key, orchestrator))
        except StopIteration:
            return None

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` can be represented by ``cls``

        .. note::

            This method should typically be overriden by all subclasses of
            ``Simple``.

        Args:
            obj (Type): the type that may be represented by ``cls``
            orchestrator (Orchestrator): container of all types, which
                is used for nested types

        Returns:
            bool: if ``True``, the type can be represented by ``cls``
        """
        return False

    def type_check(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Ensure `obj` can be converted to the type ``target_cls``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it
                can be converted to an instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        if not self.contains(target_cls, orchestrator):
            return UnRepresentedTypeError(target_cls=target_cls, unit=self, obj=obj)
        ru = self.get(target_cls, orchestrator)
        return ru._type_check(target_cls, orchestrator, obj)

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Ensure `obj` can be converted to the type ``target_cls``

        .. note::

            This method should typically be overriden by all subclasses of
            ``Simple``.

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it
                can be converted to an instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        return None

    def from_obj(self, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Convert `obj` to an instance of the type ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into an instance of
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be converted to an instance of
                ``target_cls``

        Returns:
            Any: an instance of ``target_cls`` that was derived from ``obj``

        Raises:
            UnRepresentedTypeError: ``obj`` did not map to any of the types contained in
                the ``_registry``
        """
        if not self.contains(target_cls, orchestrator):
            raise UnRepresentedTypeError(target_cls=target_cls, unit=self, obj=obj)
        ru = self.get(target_cls, orchestrator)
        return ru._from_obj(target_cls, orchestrator, obj)

    @classmethod
    def _from_obj(cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Convert `obj` to an instance of the type ``target_cls``

        .. note::

            This method should typically be overriden by all subclasses of
            ``Simple``.

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into an instance of
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be converted to an instance of
                ``target_cls``

        Returns:
            Any: an instance of ``target_cls`` that was derived from ``obj``
        """
        return None

    def relevant_types(self, key: Any, orchestrator: "Orchestrator") -> Iterable[Type]:
        """Iterable containing the types that map to the given ``key``

        Args:
            key (Any): the object to check if it maps to any types in the
                ``_registry``
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types

        Returns:
            Iterable[Type]: collection of types in the ``registry`` that map to ``key``
        """
        return (
            v_prime
            for v in self._registry.values()
            for v_prime in v.values()
            if v_prime._contains(key, orchestrator)
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['adorn',
 'adorn.alter',
 'adorn.data',
 'adorn.exception',
 'adorn.orchestrator',
 'adorn.unit',
 'adorn.unit.search']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=4.0.1,<5.0.0']

setup_kwargs = {
    'name': 'adorn',
    'version': '0.0.11',
    'description': 'Adorn',
    'long_description': 'Adorn\n======\n\n|PyPI| |Status| |Python Version| |License|\n\n|Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/adorn.svg\n   :target: https://pypi.org/project/adorn/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/adorn.svg\n   :target: https://pypi.org/project/adorn/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/adorn\n   :target: https://pypi.org/project/adorn\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/adorn\n   :target: https://opensource.org/licenses/Apache-2.0\n   :alt: License\n.. |Tests| image:: https://github.com/pyadorn/adorn/workflows/Tests/badge.svg\n   :target: https://github.com/pyadorn/adorn/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/pyadorn/adorn/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/pyadorn/adorn\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n\nFeatures\n--------\n``adorn`` is a configuration tool for python code.\n\n``adorn`` can currently\n\n* instantiate an object\n* check that a config can instantiate an object\n\n\nExample\n-------\n\n.. code-block:: python\n\n   from adorn.orchestrator.base import Base\n   from adorn.params import Params\n   from adorn.unit.complex import Complex\n   from adorn.unit.constructor_value import ConstructorValue\n   from adorn.unit.parameter_value import ParameterValue\n   from adorn.unit.python import Python\n\n\n   @Complex.root()\n   class Example(Complex):\n      pass\n\n   @Example.register(None)\n   class Parent(Example):\n       def __init__(self, parent_value: str) -> None:\n           super().__init__()\n           self.parent_value = parent_value\n\n\n   @Parent.register("child")\n   class Child(Parent):\n       def __init__(self, child_value: int, **kwargs) -> None:\n           super().__init__(**kwargs)\n           self.child_value = child_value\n\n\n   base = Base(\n       [\n           ConstructorValue(),\n           ParameterValue(),\n           Example(),\n           Python()\n       ]\n   )\n\n   params = Params(\n           {\n               "type": "child",\n               "child_value": 0,\n               "parent_value": "abc"\n           }\n   )\n\n   # well specified configuration\n   # we can type check from any level in the\n   # class hierarchy\n   assert base.type_check(Example, params) is None\n   assert base.type_check(Parent, params) is None\n   assert base.type_check(Child, params) is None\n\n   # instantiate\n   # we can instantiate from any level in the\n   # class hierarchy\n   example_obj = base.from_obj(\n       Example,\n       params\n   )\n\n   assert isinstance(example_obj, Child)\n\n\n   parent_obj = base.from_obj(\n       Parent,\n       params\n   )\n\n   assert isinstance(parent_obj, Child)\n\n\n   child_obj = base.from_obj(\n       Child,\n       params\n   )\n\n   assert isinstance(child_obj, Child)\n\n\n\nInstallation\n------------\n\nYou can install *Adorn* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install adorn\n\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `Apache 2.0 license`_,\n*Adorn* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_\'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _Apache 2.0 license: https://opensource.org/licenses/Apache-2.0\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/pyadorn/adorn/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://adorn.readthedocs.io/en/latest/usage.html\n',
    'author': 'Jacob Baumbach',
    'author_email': 'jacob.baumbach@hey.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pyadorn/adorn',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""State from the constructor for a parameter from the constructor."""
from typing import Any
from typing import Dict
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from adorn.unit.complex import _UnitT


class Parameter:
    """State for a parameter from a given :class:`~adorn.data.constructor.Constructor`

    ``Parameter`` allows for information outside of its ``obj``, to be utilized,
    to perform some action.  Examples of constructs that utilize the information
    contained in ``Parameter`` include:

    - :class:`~adorn.unit.parameter_value.DependentTypeCheck`
    - :class:`~adorn.unit.parameter_value.DependentFromObj`
    - :class:`~adorn.unit.parameter_value.DependentUnion`

    Attributes:
        cls (_UnitT): the type of the underlying parameter
        parent (_UnitT): The type of object whose constructor requires
            an object of type ``cls``
        local_state (Dict[str, Any]): Information about other arguments
            provided to the given constructor
        parameter_name (str): name of the parameter in the constructor
    """

    def __init__(
        self,
        cls: "_UnitT",
        parent: "_UnitT",
        local_state: Dict[str, Any],
        parameter_name: str,
    ):
        self.cls = cls
        self.parent = parent
        self.local_state = local_state
        self.parameter_name = parameter_name
        self.origin = getattr(self.cls, "__origin__", None)
        self.args = getattr(self.cls, "__args__", None)

    def __eq__(self, o: object) -> bool:  # noqa: C901
        # Literal.__eq__ doesn't support dict's so
        # we force the check
        # prevent circular import
        from adorn.unit.parameter_value import Dependent

        if not isinstance(o, Parameter):
            return False

        normal_args = all(
            i
            for i in [
                self.parent == o.parent,
                self.local_state == o.local_state,
                self.parameter_name == o.parameter_name,
            ]
        )
        if not normal_args:
            return False

        eq_cls = False
        if issubclass(getattr(self.cls, "__origin__", int), Dependent):
            eq_cls = str(self.cls) == str(o.cls)
        else:
            eq_cls = self.cls == o.cls
        return eq_cls

    def __str__(self):
        return str(self.cls)
